/*
 * lzwscan.c
 *
 * Copyright (c) 2002 Marius Aamodt Eriksen <marius@monkey.org>
 *
 * $Id: lzwscan.c,v 1.5 2003/08/09 04:58:13 marius Exp $
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <assert.h>
#include <err.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <zlib.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif /* HAVE_STDINT_H */

#include "netics.h"

#define ZERR(z)   zerr((z), __FILE__, __LINE__)
#define ZWARN(z)  zwarn((z), __FILE__, __LINE__)
#define ZBUFSZ(x) (ceil(1.01 * (x)))

struct lzw {
	z_stream   z;
	u_char    *buf;
	uint32_t   buflen;
	int        precise;
};

void   zerr(z_stream *, char *, int);
void   zwarn(z_stream *, char *, int);
double result(struct lzw *);
long   makeprecise(struct lzw *);

void *
lzw_new(uint32_t buflen, int options)
{
	struct lzw *l;

	l = calloc(1, sizeof(*l));
	assert(l != NULL);

	l->z.zalloc = Z_NULL;
	l->z.zfree = Z_NULL;
	l->z.avail_in = 0;
	l->z.next_in = NULL;	

	if  (deflateInit(&l->z, Z_BEST_COMPRESSION) != Z_OK)
		ZERR(&l->z);

	l->buf = (u_char *)malloc(ZBUFSZ(buflen)); 
	l->buflen = ZBUFSZ(buflen);
	assert(l->buf != NULL);

	if (options & OPT_PRECISE)
		l->precise = 1;
	else
		l->precise = 0;

	return (l);
}

void
lzw_free(void *v)
{
	struct lzw *l = (struct lzw *)v;

	free(l);
}

uint32_t
lzw_feed(void *v, u_char *buf, uint32_t buflen)
{
	struct lzw *l = (struct lzw *)v;

	l->z.avail_in = buflen;
	l->z.next_in = buf;
	l->z.next_out = l->buf;
	l->z.avail_out = ZBUFSZ(l->buflen);
	
	if (deflate(&l->z, Z_NO_FLUSH) != Z_OK)
		ZERR(&l->z);

	/* we can always return a result */
	return (1);
}

double
lzw_result_raw(void *v)
{
	struct lzw *l = (struct lzw *)v;
	
	return result(l);
}

double
lzw_result_percent(void *v)
{
	struct lzw *l = (struct lzw *)v;
	double val;

	val = result(l);

	return (1 / val);
}


double
result(struct lzw *l)
{
	long out_len;
	double val = 0.0;

	/* cache results, use total_out + total_in as generation number */

        if (l->precise)
                out_len = makeprecise(l);
        else
                out_len = l->z.total_out;
        
        if (out_len != 0 && l->z.total_in != 0)
                val = (1.0 * l->z.total_in) / (1.0 * out_len);
	
	return (val);
}

long
makeprecise(struct lzw *l)
{
	z_stream *z = &l->z;
	z_stream zview;
	int zstatus;
	long ret;
	char *out;

 	out = (char *)malloc(ZBUFSZ(l->buflen));

	assert(out != NULL);
	
	if (deflateCopy(&zview, z) != Z_OK)
		ZERR(z);

	zview.avail_out = ZBUFSZ(l->buflen);
	zview.next_out = out;

	while ((zstatus = deflate(&zview, Z_FINISH)) == Z_OK) {
		zview.avail_out = ZBUFSZ(l->buflen);
		zview.next_out = out;
	}
	
	if (zstatus != Z_STREAM_END) {
		warnx("precise measurement error");
		return (0);
	}

	ret = zview.total_out;
	free(out);
	deflateEnd(&zview);	

	return (ret);
}

void
lzw_reset(void *v)
{
	struct lzw *l = (struct lzw *)v;
	
	if (deflateReset(&l->z) != Z_OK)
		ZWARN(&l->z);
}

void
zerr(z_stream *z, char *file, int line)
{
	if (z->msg != NULL)
		errx(1, "%s:%i compression error: %s", file, line, z->msg);
	else
		errx(1, "%s:%i compression error", file, line);
}

void
zwarn(z_stream *z, char *file, int line)
{
	if (z->msg != NULL)
		warnx("%s:%i compression warning: %s", file, line, z->msg);
	else
		warnx("%s:%i compression warning", file, line);
}

struct sc_ops lzw_ops = {
	name:              "lzw",
	sc_new:            lzw_new,
	sc_free:           lzw_free,
	sc_feed:           lzw_feed,
	sc_reset:          lzw_reset,
	sc_result_raw:     lzw_result_raw,
	sc_result_percent: lzw_result_percent,
	sc_result_display: NULL
};
