/*
 * maurerscan.c
 *
 * Copyright (c) 2002 Marius Aamodt Eriksen <marius@monkey.org>
 * All rights reserved.
 *
 * $Id: maurerscan.c,v 1.3 2003/05/01 08:21:15 marius Exp $
 */

/*
 * This implementation of the Ueli Maurer's universal statistical test
 * algorithm was adopted from
 *    http://www.geocities.com/SiliconValley/Code/4704/uliscanc.txt
 */

/*
 * XXX: better handling when result is ready.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif /* HAVE_STDINT_H */

#include "netics.h"

#define L 8
#define V (1 << L)
#define Q (10 * V)
#define K (100 * Q)

#define MAURER_EXPECTED 7.1836656

struct maurer {
	uint32_t  i;
	uint32_t  j;
	double    sum;
	double    result;
	int       table[V];
};

void * 
maurer_new(uint32_t buflen, int options)
{
	struct maurer *ret;

	ret = malloc(sizeof(*ret));
	if (ret == NULL)
		err(1, "malloc");

	ret->i = 0;
	ret->j = Q;
	ret->sum = 0.0;
	ret->result = -1;
	memset(ret->table, 0, sizeof(ret->table));

	return ((void *)ret);
}

void
maurer_free(void *v)
{
	struct maurer *m = v;

	free(m);
}

/* XXX this needs to be changed so that we don't loose bytes */
uint32_t
maurer_feed(void *v, u_char *buf, uint32_t buflen)
{
	struct maurer *m = v;
	uint32_t k = 0, q;
	
	/* Fill initial table */
	if (m->i < Q) {
		for (; k < buflen && m->i < Q; m->i++, k++)
			m->table[buf[k]] = m->i;
		
		/* Return negative for more bytes needed */
		if (m->i < Q)
			return -(Q - m->i);
	}
	
	for (; m->j < Q + K && k < buflen; m->j++, k++) {
		q = m->j;
		if (m->table[buf[k]] > q)
			q += K;
		m->sum += log((double)(m->j - m->table[buf[k]]));
		m->table[buf[k]] = m->j;
	}

	/*
	 * Return non-negative when we have excess bytes, and a result
	 * is available
	 */
	if (m->j == Q + K) {
		m->j = Q;
		m->result = (m->sum / (double)(K)) / log(2.0);
		m->sum = 0.0;
		return (buflen - k);
	}
	
	/* Return 0 if we can yield a result */
	return (0);
}

void
maurer_reset(void *v)
{
	struct maurer *m = v;

	memset(m->table, 0, sizeof(m->table));
	m->j = Q;
	m->i = 0;
	m->sum = 0.0;
	m->result = -1.0;
}

double
maurer_result_percent(void *v)
{
	struct maurer *m = v;

	return (m->result / MAURER_EXPECTED);
}

double
maurer_result_raw(void *v)
{
	struct maurer *m = v;

	return (m->result);
}

struct sc_ops maurer_ops = {
	.name              = "maurer",
	.sc_new            = maurer_new,
	.sc_free           = maurer_free,
	.sc_feed           = maurer_feed,
	.sc_reset          = maurer_reset,
	.sc_result_raw     = maurer_result_raw,
	.sc_result_percent = maurer_result_percent,
	.sc_result_display = NULL
};
