/*
 * netics.c
 *
 * Copyright (c) 2002, 2003 Marius Aamodt Eriksen <marius@monkey.org>
 * All rights reserved.
 *
 * $Id: netics.c,v 1.4 2003/08/09 04:58:13 marius Exp $
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/tree.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif /* HAVE_SYS_TIME_H */

#include <assert.h>
#include <err.h>
#include <event.h>
#include <libnet.h>
#include <math.h>
#include <nids.h>
#include <pcap.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#if defined(HAVE_TIME_H) && defined(TIME_WITH_SYS_TIME)
#include <time.h>
#endif /* defined(HAVE_TIME_H) && defined(TIME_WITH_SYS_TIME) */
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif /* HAVE_STDINT_H */

#include "netics.h"
#include "pcapu.h"
#include "atomicio.h"

#define NIDS_POLL_TIMEOUT {0, 10000}
#define SNAPLEN 1540

#define DIFF(a,b) do {				\
        if ((a) < (b)) return (-1);		\
        if ((a) > (b)) return (1);		\
} while (0)

struct conn {
	SPLAY_ENTRY(conn)  node;
	struct tuple4      tuple;	
	struct sc_ops     *ops;
	void              *handle;
};

static int conn_compare(struct conn *, struct conn *);

SPLAY_HEAD(conn_tree, conn);
SPLAY_PROTOTYPE(conn_tree, conn, node, conn_compare);
SPLAY_GENERATE(conn_tree, conn, node, conn_compare);

static int conf_lastint, conf_numint,
    conf_timeout, conf_progressbar, conf_sc_opt, conf_verbose;
static struct sc_ops *conf_ops;
static void *conf_handle;

static struct event timeout_ev, nids_ev, sigint_ev;
extern struct sc_ops maurer_ops, lzw_ops;
static struct sc_ops *scans[] = { &lzw_ops, &maurer_ops, NULL };
static struct conn_tree conn_tree_head;
static struct timeval nids_tv = NIDS_POLL_TIMEOUT;

static void usage(void);
static void checkpoint(void);
static void evsig_cb(int, short, void *);
static void timeout_cb(int, short, void *);
static void updatebar(double);
static void xsyslog(int, int, void *, void *);
static void tcp_cb(struct tcp_stream *ts, void **data);
static void tcp_half_cb(struct tuple4 *, struct half_stream *, struct conn *, int);
static void udp_cb(struct tuple4 *, u_char *, int, struct ip *);
static void nids_cb(int fd, short event, void *arg);

int
main(int argc, char **argv)
{
	int opt;
#ifndef __APPLE__
	int nids_fd;
#endif
	char *file = NULL, *filter = "tcp or udp", *scan = "lzw";
	extern char *__progname;
	struct timeval *xnids_tv = NULL, timeout_tv;
	struct sc_ops **scani;

	conf_sc_opt = OPT_PRECISE;
	conf_timeout = 10;

	event_init();

	while ((opt = getopt(argc, argv, "phVbvi:r:t:s:")) != -1)
 		switch (opt) {
		case 'i':
			nids_params.device = optarg;
			break;
		case 'P':
			conf_sc_opt &= ~OPT_PRECISE;
			break;
		case 'r':
			file = strdup(optarg);
			break;
		case 't':
			conf_timeout = atoi(optarg);
			break;
		case 'b':
			conf_progressbar = 1;
			break;
		case 's':
			scan = strdup(optarg);
			break;
		case 'v':
			conf_verbose++;
			break;
		case 'V':
			fprintf(stderr, "%s version " VERSION "\n", __progname);
			exit(0);
		case 'h':
		default:
			usage();
			/* NOTREACHED */
		}

	argc -= optind;
	argv += optind;

	scani = scans;

	while (*scani != NULL) {
		if (!strcmp((*scani)->name, scan)) {
			conf_ops = *scani;
			break;
		}
		scani++;
	}

	if (*scani == NULL)
		errx(1, "Could not find scan %s", scan);

	if (conf_verbose > 0)
		warnx("Using %s scan", conf_ops->name);

	conf_handle = (*conf_ops->sc_new)(SNAPLEN, conf_sc_opt);

	nids_params.pcap_filter = argc > 0 ? copy_argv(argv) : filter;
	nids_params.syslog = xsyslog;

	if (!nids_init())
		errx(1, "nids_init: %s", nids_errbuf);

	nids_register_tcp(tcp_cb);
	nids_register_udp(udp_cb);

#ifndef __APPLE__
	if ((nids_fd = nids_getfd()) != -1)
		event_set(&nids_ev, nids_fd, EV_READ, nids_cb, NULL);
	else {
		warnx("Could not get file descriptor: %s.  Polling instead.",
		    nids_errbuf);
#endif /* !__APPLE__ */
		xnids_tv = &nids_tv;
		evtimer_set(&nids_ev, nids_cb, NULL);
#ifndef __APPLE__
	}
#endif

	event_add(&nids_ev, xnids_tv);

	event_set(&sigint_ev, SIGINT, EV_SIGNAL | EV_PERSIST,
	    evsig_cb, NULL);
	event_add(&sigint_ev, NULL);

	SPLAY_INIT(&conn_tree_head);

	if (conf_timeout > 0) {
		evtimer_set(&timeout_ev, timeout_cb, NULL);
		timerclear(&timeout_tv);
		evtimer_add(&timeout_ev, &timeout_tv);
	} else
		errx(1, "Invalid timeout value");

	warnx("Listening on %s [%s]", nids_params.device, 
	    nids_params.pcap_filter);

	event_dispatch();

	/* NOTREACHED */
	return (0);
}

static void
nids_cb(int fd, short event, void *arg)
{
	struct timeval *xtv = NULL;

	if (nids_next() == -1)
		warnx("nids_next(): %s", nids_errbuf);

	if (event == EV_TIMEOUT)
		xtv = &nids_tv;

	event_add(&nids_ev, xtv);
}

static void
tcp_cb(struct tcp_stream *ts, void **data)
{
	struct half_stream *tsc = &ts->client, *tss = &ts->server;
	struct conn *c = *data;
	void *handle;

	switch (ts->nids_state) {
	case NIDS_JUST_EST:
		if (conf_verbose > 1)
			warnx("TCP NEW %s:%d <-> %s:%d",
			    libnet_host_lookup(ts->addr.saddr, 1),
			    ntohs(ts->addr.source),
			    libnet_host_lookup(ts->addr.daddr, 1),
			    ntohs(ts->addr.dest)); 

		tss->collect = tsc->collect = 1;
		handle = (*conf_ops->sc_new)(SNAPLEN, conf_sc_opt);

		c = calloc(1, sizeof(*c));
		if (c ==  NULL) {
			*data = NULL;
			return;
		}
		c->handle = handle;
		c->ops = conf_ops;
		memcpy(&c->tuple, &ts->addr, sizeof(c->tuple));
		SPLAY_INSERT(conn_tree, &conn_tree_head, c);
		*data = c;

		break;
	case NIDS_DATA:
		if (c == NULL)
			return;

		if (tss->count_new > 0)
			tcp_half_cb(&ts->addr, tss, c, 1);
		if (tsc->count_new > 0)
			tcp_half_cb(&ts->addr, tsc, c, 0);

		break;
	default:
		if (conf_verbose > 1)
			warnx("TCP STALE %s:%d <-> %s:%d",
			    libnet_host_lookup(ts->addr.saddr, 1),
			    ntohs(ts->addr.source),
			    libnet_host_lookup(ts->addr.daddr, 1),
			    ntohs(ts->addr.dest)); 

		SPLAY_REMOVE(conn_tree, &conn_tree_head, c);

		free(c);
		break;
	}
}

static void
tcp_half_cb(struct tuple4 *addr, struct half_stream *hs, struct conn *c,
    int serv)
{
	u_char *buf;
	int buflen;
	uint32_t saddr, daddr;
	uint16_t sport, dport;

	buf = hs->data;
	buflen = hs->count_new;

	saddr = serv ? addr->saddr : addr->daddr;
	daddr = serv ? addr->daddr : addr->saddr;
	sport = serv ? addr->source : addr->dest;
	dport = serv ? addr->dest : addr->source;

	if (conf_verbose > 2)
		warnx("TCP DATA %s:%d -> %s:%d (%d)",
		    libnet_host_lookup(saddr, 1),
		    ntohs(sport),
		    libnet_host_lookup(daddr, 1),
		    ntohs(dport), buflen);

	if (buf == NULL || buflen <= 0)
		return;

	(*c->ops->sc_feed)(c->handle, buf, buflen);
	/* Global */
	(*conf_ops->sc_feed)(conf_handle, buf, buflen);
}

static void
udp_cb(struct tuple4 *tuple, u_char *buf, int buflen, struct ip *iph)
{
	if (conf_verbose > 2)
		warnx("UDP DATA %s:%d -> %s:%d (%d)",
		    libnet_host_lookup(tuple->saddr, 1),
		    ntohs(tuple->source),
		    libnet_host_lookup(tuple->daddr, 1),
		    ntohs(tuple->dest), buflen);

	(*conf_ops->sc_feed)(conf_handle, buf, buflen);
}

static void
checkpoint(void)
{
	double val;

	if (conf_ops->sc_result_display != NULL) {
		(*conf_ops->sc_result_display)(conf_handle);
		return;
	}

	if (conf_progressbar) {
		if ((val = (conf_ops->sc_result_percent)(conf_handle)) < 0)
			goto nodata;
		updatebar(val);
	} else {
		if ((val = (conf_ops->sc_result_raw)(conf_handle)) < 0) 
			goto nodata;
		printf("%d -> %f\n", (int)time(NULL), val);
	}

	return;

 nodata:
	printf("\rdata not yet available");
	fflush(stdout);
}

static void
evsig_cb(int fd, short ev, void *arg)
{
	long curtime;

	if (conf_progressbar)
		exit(0);

	checkpoint();

	curtime = (int)time(NULL);

	if (curtime - conf_lastint < 2)
		conf_numint++;
	else
		conf_numint = 0;

	if (conf_numint == 2)
		exit(0);

	conf_lastint = curtime;
}

static void
timeout_cb(int fd, short event, void *arg)
{
	struct timeval tv;

	checkpoint();

	timerclear(&tv);
	tv.tv_sec = conf_timeout;
	event_add(&timeout_ev, &tv);
}

#define BUFSZ 256
static void
updatebar(double value)
{
	uint32_t width, progress, i;
	struct winsize winsz;
	char buf[BUFSZ];
	static int barpos = 0;
	static char *bars = "|/-\\";

	if (value < 0) {
		printf("\rdata not yet available");
		fflush(stdout);
		return;
	}
		
	if (ioctl(fileno(stdout), TIOCGWINSZ, &winsz) != -1)
		width = winsz.ws_col && winsz.ws_col < 256 ? winsz.ws_col : 80;
	else
		width = 80;
	
	width -= 10;
	
	if (value < 0) 
		value = 0.0;
	else if (value > 1.0) 
		value = 1.0;
	
	progress = ceil(value * width);
	
	snprintf(buf, sizeof(buf), "\r%3d%% |", (int)ceil(100 * value));
	
	i = strlen(buf);
	if (i + width > BUFSZ)
		return;
	memset(buf + i, '*', progress); i += progress;
	memset(buf + i, ' ', width - progress); i += width - progress;
	buf[i] = '\0';

	/* XXX do integers overflow on a multiple of 4? :) */
	snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), " |%c",
	    bars[barpos++ % sizeof(bars)]);
	
	atomicio(write, fileno(stdout), buf, strlen(buf));
}
#undef BUFSZ /* 256 */

static int
conn_compare(struct conn *a, struct conn *b)
{
	DIFF(a->tuple.saddr, b->tuple.saddr);
	DIFF(a->tuple.daddr, b->tuple.daddr);
	DIFF(a->tuple.source, b->tuple.source);
	DIFF(a->tuple.dest, b->tuple.dest);
	/* DIFF(a->tuple.proto, b->tuple.proto); */

	/* Equal */
	return (0);
}

static void
xsyslog(int type, int errnum, void *iph, void *data)
{}

static void
usage(void)
{
	extern char *__progname;
	struct sc_ops **scani;

	fprintf(stderr,
	    "%s: [-phVbv] [-i <interface>] [-r <file>] [-s <scan>] "
	    "[-t <timeout>] [filter]\n"
	    "\t-i <interface> Listen on specified interface\n"
	    "\t-r <file>      Use tcpdump file <file> for sniffing\n"
	    "\t-t <timeout>   Checkpoint every <timeout> seconds\n"
	    "\t-s <scan>      Use scan <scan>.  available scans are:\n",
	    __progname);
	scani = scans;
	fprintf(stderr,
	    "\t               %s (default)", (*scani++)->name);
	while (*scani != NULL) {
		fprintf(stderr, ", %s", (*scani)->name);
		scani++;
	}
	fprintf(stderr, 	    
	    "\n"
	    "\t-P             Use more faster (more imprecise) measurements\n"
	    "\t-h             Help\n"
	    "\t-V             Print version number\n"
	    "\t-b             Show measurement using a \"progress\" bar\n"
	    "\t-v             Increase verbosity level\n");

	exit(1);
}

